__all__ = ['system_packages']

system_packages = [
    "docassemble",
    "docassemble.base",
    "docassemble.demo",
    "3to2",
    "acme",
    "airtable-python-wrapper",
    "alembic",
    "amqp",
    "asn1crypto",
    "astunparse",
    "async-generator",
    "async-timeout",
    "atomicwrites",
    "attrs",
    "azure-common",
    "azure-core",
    "azure-identity",
    "azure-keyvault-secrets",
    "azure-nspkg",
    "azure-storage-blob",
    "Babel",
    "backports.zoneinfo",
    "bcrypt",
    "beautifulsoup4",
    "behave",
    "bidict",
    "billiard",
    "bleach",
    "blinker",
    "boto",
    "boto3",
    "botocore",
    "cachetools",
    "cairocffi",
    "CairoSVG",
    "celery",
    "certbot",
    "certbot-apache",
    "certbot-nginx",
    "certifi",
    "cffi",
    "chardet",
    "charset-normalizer",
    "click",
    "click-didyoumean",
    "click-plugins",
    "click-repl",
    "clicksend-client",
    "colorama",
    "commonmark",
    "ConfigArgParse",
    "configobj",
    "configparser",
    "convertapi",
    "crayons",
    "cryptography",
    "cssselect2",
    "da-pkg-resources",
    "defusedxml",
    "Deprecated",
    "distro",
    "dnspython",
    "docassemble-backports",
    "Docassemble-Flask-User",
    "Docassemble-Pattern",
    "docassemble-textstat",
    "docassemblekvsession",
    "docopt",
    "docutils",
    "docxcompose",
    "docxtpl",
    "email-validator",
    "et-xmlfile",
    "eventlet",
    "Flask",
    "Flask-Babel",
    "Flask-Cors",
    "Flask-Login",
    "Flask-Mail",
    "Flask-SocketIO",
    "Flask-SQLAlchemy",
    "Flask-WTF",
    "future",
    "gcs-oauth2-boto-plugin",
    "geographiclib",
    "geopy",
    "google-api-core",
    "google-api-python-client",
    "google-auth",
    "google-auth-httplib2",
    "google-auth-oauthlib",
    "google-cloud-core",
    "google-cloud-storage",
    "google-cloud-translate",
    "google-cloud-vision",
    "google-crc32c",
    "google-i18n-address",
    "google-reauth",
    "google-resumable-media",
    "googleapis-common-protos",
    "greenlet",
    "grpcio",
    "grpcio-status",
    "gspread",
    "guess-language-spirit",
    "h11",
    "httplib2",
    "humanize",
    "Hyphenate",
    "idna",
    "importlib-metadata",
    "importlib-resources",
    "iniconfig",
    "iso8601",
    "isodate",
    "itsdangerous",
    "jdcal",
    "jeepney",
    "jellyfish",
    "Jinja2",
    "jmespath",
    "joblib",
    "josepy",
    "keyring",
    "kombu",
    "libcst",
    "links-from-link-header",
    "lxml",
    "Mako",
    "Markdown",
    "MarkupSafe",
    "mdx-smartypants",
    "minio",
    "mod_wsgi",
    "mod-wsgi",
    "monotonic",
    "msal",
    "msal-extensions",
    "msrest",
    "mypy-extensions",
    "namedentities",
    "netifaces",
    "nltk",
    "num2words",
    "numpy",
    "oauth2client",
    "oauthlib",
    "openpyxl",
    "ordered-set",
    "outcome",
    "packaging",
    "pandas",
    "parse",
    "parse-type",
    "parsedatetime",
    "passlib",
    "pathlib",
    "pdfminer.six",
    "phonenumbers",
    "Pillow",
    "pip",
    "pkginfo",
    "pluggy",
    "ply",
    "portalocker",
    "prompt-toolkit",
    "proto-plus",
    "protobuf",
    "psutil",
    "psycopg2-binary",
    "py",
    "pyasn1",
    "pyasn1-modules",
    "pycountry",
    "pycparser",
    "pycryptodome",
    "pycryptodomex",
    "pycurl",
    "Pygments",
    "PyJWT",
    "PyLaTeX",
    "PyNaCl",
    "PyOpenSSL",
    "pyOpenSSL",
    "pyotp",
    "pyparsing",
    "pyPdf",
    "PyPDF2",
    "pypdftk",
    "pypng",
    "pyRFC3339",
    "PySocks",
    "pytest",
    "python-augeas",
    "python-dateutil",
    "python-docx",
    "python-editor",
    "python-engineio",
    "python-http-client",
    "python-ldap",
    "python-socketio",
    "pytz",
    "pytz-deprecation-shim",
    "pyu2f",
    "PyYAML",
    "pyzbar",
    "qrcode",
    "rauth",
    "readme-renderer",
    "redis",
    "regex",
    "reportlab",
    "repoze.lru",
    "requests",
    "requests-oauthlib",
    "requests-toolbelt",
    "retry-decorator",
    "rfc3339",
    "rfc3986",
    "rich",
    "rsa",
    "ruamel.yaml",
    "ruamel.yaml.clib",
    "s3transfer",
    "s4cmd",
    "scikit-learn",
    "scipy",
    "SecretStorage",
    "selenium",
    "sendgrid",
    "setuptools",
    "simplekv",
    "six",
    "sklearn",
    "sniffio",
    "SocksiPy-branch",
    "sortedcontainers",
    "soupsieve",
    "SQLAlchemy",
    "starkbank-ecdsa",
    "tailer",
    "telnyx",
    "threadpoolctl",
    "tinycss2",
    "titlecase",
    "toml",
    "tomli",
    "tqdm",
    "trio",
    "trio-websocket",
    "twilio",
    "twine",
    "typing-extensions",
    "typing-inspect",
    "tzdata",
    "tzlocal",
    "ua-parser",
    "uritemplate",
    "urllib3",
    "us",
    "user-agents",
    "uWSGI",
    "vine",
    "wcwidth",
    "webdriver-manager",
    "webencodings",
    "Werkzeug",
    "wheel",
    "wrapt",
    "wsproto",
    "WTForms",
    "xfdfgen",
    "xlrd",
    "XlsxWriter",
    "xlwt",
    "zipp",
    "zope.component",
    "zope.event",
    "zope.hookable",
    "zope.interface"
]
